#brailleDisplayDrivers/braillesense.py
#A part of NonVisual Desktop Access (NVDA)
#This file is covered by the GNU General Public License.
#See the file COPYING for more details.
#Copyright (C) 2010 Gianluca Casalino <gianluca@spazioausili.net>
import speech
import wx
import tones
from ctypes import *
from ctypes.wintypes import *
import braille
from logHandler import log
import inputCore
pressedKeys=set()
releasedKeys=set()
BS_KEYPRESSED = 0x01
BS_KEYRELEASED = 0x02
BS_CURSORROUTING = 0x00
LRESULT=c_long
HCURSOR=c_long

#MAP OF KEYS
BS_KEYS = {
	1: 'Dot1',
	2: 'Dot2',
	4: 'Dot3',
	8: 'Dot4',
	16: 'Dot5',
	32: 'Dot6',
	64: 'Dot7',
	128: 'Dot8',
	131072: 'scroll_left',
	65536: 'scroll_up',
	262144: 'scroll_down',
	524288: 'scroll_right',
	256: 'space',
	512: 'advance1',
	1024: 'advance2',
	2048: 'advance3',
	4096: 'advance4',
	8192: 'scroll_left',
	16384: 'scroll_right'}
_ignoreKeyReleases = False

bsLib = cdll.LoadLibrary("brailleDisplayDrivers\\HanSoneConnect.dll")

#Standard window class stuff

WNDPROC=WINFUNCTYPE(LRESULT,HWND,c_uint,WPARAM,LPARAM)

class WNDCLASSEXW(Structure):
	_fields_=[
		('cbSize',c_uint),
		('style',c_uint),
		('lpfnWndProc',WNDPROC),
		('cbClsExtra',c_int),
		('cbWndExtra',c_int),
		('hInstance',HINSTANCE),
		('hIcon',HICON),
		('HCURSOR',HCURSOR),
		('hbrBackground',HBRUSH),
		('lpszMenuName',LPWSTR),
		('lpszClassName',LPWSTR),
		('hIconSm',HICON),
	]

appInstance=windll.kernel32.GetModuleHandleW(None)

nvdaBsBrlWm=windll.user32.RegisterWindowMessageW(u"nvdaBsBrlWm")

@WNDPROC
def nvdaBsBrlWndProc(hwnd,msg,wParam,lParam):
	global pressedKeys, releasedKeys, _ignoreKeyReleases
	if msg == nvdaBsBrlWm and wParam==BS_KEYRELEASED:
		if not _ignoreKeyReleases and pressedKeys:
			try: 
				inputCore.manager.executeGesture(InputGesture(pressedKeys))
			except inputCore.NoInputGestureAction:
				pass
			_ignoreKeyReleases = True
		pressedKeys.discard(lParam)
	elif msg == nvdaBsBrlWm and wParam == BS_CURSORROUTING:
		try:
			inputCore.manager.executeGesture(InputGesture(lParam))
		except inputCore.NoInputGestureAction:
			pass		
	elif msg == nvdaBsBrlWm and  wParam == BS_KEYPRESSED:
		pressedKeys.add(lParam)
		log.info(lParam)
		_ignoreKeyReleases = False
	return windll.user32.DefWindowProcW(hwnd,msg,wParam,lParam)

nvdaBsBrlWndCls=WNDCLASSEXW()
nvdaBsBrlWndCls.cbSize=sizeof(nvdaBsBrlWndCls)
nvdaBsBrlWndCls.lpfnWndProc=nvdaBsBrlWndProc
nvdaBsBrlWndCls.hInstance=appInstance
nvdaBsBrlWndCls.lpszClassName=u"nvdaBsBrlWndCls"

class BrailleDisplayDriver(braille.BrailleDisplayDriver):
	"""Braille Sense.
	"""
	name = "braillesense"
	description = _("Braille Sense")

	@classmethod
	def check(cls):
		return True #bool(bsLib)

	def __init__(self):
		super(BrailleDisplayDriver, self).__init__()
		self._messageWindowClassAtom=windll.user32.RegisterClassExW(byref(nvdaBsBrlWndCls))
		self._messageWindow=windll.user32.CreateWindowExW(0,self._messageWindowClassAtom,u"nvdaBsBrlWndCls window",0,0,0,0,0,None,None,appInstance,None)
		connected = bsLib.Open("usb",self._messageWindow,nvdaBsBrlWm)
		if connected==0: 
			for n in xrange(1,256):
				port=("com%d"%n)
				if n > 8: port="\\\\.\\"+port
				connected = bsLib.Open(port,self._messageWindow,nvdaBsBrlWm)
				if connected>=1: return
			raise RuntimeError("No display found")
 
	def terminate(self):
		super(BrailleDisplayDriver, self).terminate()
		bsLib.Close()
		windll.user32.DestroyWindow(self._messageWindow)
		windll.user32.UnregisterClassW(self._messageWindowClassAtom,appInstance)

 	def _get_numCells(self):
		return bsLib.GetBSCellCount()

	def display(self, cells):
		cells = "".join([chr(x) for x in cells])
		bsLib.SendData(cells)
	gestureMap = inputCore.GlobalGestureMap({
		"globalCommands.GlobalCommands": {
			"kb:space": ("br(braillesense):space",),
			"kb:leftAlt": ("br(braillesense):Dot1+Dot3+Dot4+space",),
			"kb:capsLock": ("br(braillesense):Dot1+Dot3+Dot6+space",),
			"kb:tab": ("br(braillesense):Dot4+Dot5+space",),
			"kb:shift+alt+tab": ("br(braillesense):advance2+advance3+advance1",),
			"kb:alt+tab": ("br(braillesense):advance2+advance3",),
			"kb:shift+tab": ("br(braillesense):Dot1+Dot2+space",),
			"kb:end": ("br(braillesense):Dot4+Dot6+space",),
			"kb:control+end": ("br(braillesense):Dot4+Dot5+dot6+space",),
			"kb:home": ("br(braillesense):Dot1+Dot3+space",),
			"kb:control+home": ("br(braillesense):Dot1+Dot2+dot3+space",),
			"kb:leftArrow": ("br(braillesense):Dot3+space",),
			"kb:control+shift+leftArrow": ("br(braillesense):Dot2+Dot8+space+advance1",),
			"kb:control+leftArrow": ("br(braillesense):Dot2+space",),
			"kb:shift+alt+leftArrow": ("br(braillesense):Dot2+Dot7+advance1",),
			"kb:alt+leftArrow": ("br(braillesense):Dot2+Dot7",),
			"kb:rightArrow": ("br(braillesense):Dot6+space",),
			"kb:control+shift+rightArrow": ("br(braillesense):Dot5+Dot8+space+advance1",),
			"kb:control+rightArrow": ("br(braillesense):Dot5+space",),
			"kb:shift+alt+rightArrow": ("br(braillesense):Dot5+Dot7+advance1",),
			"kb:alt+rightArrow": ("br(braillesense):Dot5+Dot7",),
			"kb:pageUp": ("br(braillesense):Dot1+Dot2+Dot6+space",),
			"kb:control+pageUp": ("br(braillesense):Dot1+Dot2+Dot6+Dot8+space",),
			"kb:upArrow": ("br(braillesense):Dot1+space",),
			"kb:control+shift+upArrow": ("br(braillesense):Dot2+Dot3+Dot8+space+advance1",),
			"kb:control+upArrow": ("br(braillesense):Dot2+Dot3+space",),
			"kb:shift+alt+upArrow": ("br(braillesense):Dot2+Dot3+Dot7+advance1",),
			"kb:alt+upArrow": ("br(braillesense):Dot2+Dot3+Dot7",),
			"kb:shift+upArrow": ("br(braillesense):scroll_left+space",),
			"kb:pageDown": ("br(braillesense):Dot3+Dot4+Dot5+space",),
			"kb:control+pageDown": ("br(braillesense):Dot3+Dot4+Dot5+Dot8+space",),
			"kb:downArrow": ("br(braillesense):Dot4+space",),
			"kb:control+shift+downArrow": ("br(braillesense):Dot5+Dot6+Dot8+space+advance1",),
			"kb:control+downArrow": ("br(braillesense):Dot5+Dot6+space",),
			"kb:shift+alt+downArrow": ("br(braillesense):Dot5+Dot6+Dot7+advance1",),
			"kb:alt+downArrow": ("br(braillesense):Dot5+Dot6+Dot7",),
			"kb:shift+downArrow": ("br(braillesense):space+scroll_right",),
			"kb:backspace": ("br(braillesense):Dot7",),
			"kb:enter": ("br(braillesense):Dot8",),
			"kb:escape": ("br(braillesense):Dot1+Dot5+space",),
			"kb:delete": ("br(braillesense):Dot1+Dot3+Dot5+space",),
			"kb:f1": ("br(braillesense):Dot1+Dot2+Dot5+space",),
			"kb:f3": ("br(braillesense):Dot1+Dot2+Dot4+Dot8",),
			"kb:f4": ("br(braillesense):Dot7+advance3",),
			"kb:a": ("br(braillesense):Dot1",),
			"kb:b": ("br(braillesense):Dot1+Dot2",),
			"kb:c": ("br(braillesense):Dot1+Dot4",),
			"kb:d": ("br(braillesense):Dot1+Dot4+Dot5",),
			"kb:e": ("br(braillesense):Dot1+Dot5",),
			"kb:f": ("br(braillesense):Dot1+Dot2+Dot4",),
			"kb:g": ("br(braillesense):Dot1+Dot2+Dot4+Dot5",),
			"kb:h": ("br(braillesense):Dot1+Dot2+Dot5",),
			"kb:i": ("br(braillesense):Dot2+Dot4",),
			"kb:j": ("br(braillesense):Dot2+Dot4+Dot5",),
			"kb:k": ("br(braillesense):Dot1+Dot3",),
			"kb:l": ("br(braillesense):Dot1+Dot2+Dot3",),
			"kb:m": ("br(braillesense):Dot1+Dot3+Dot4",),
			"kb:n": ("br(braillesense):Dot1+Dot3+Dot4+Dot5",),
			"kb:o": ("br(braillesense):Dot1+Dot3+Dot5",),
			"kb:p": ("br(braillesense):Dot1+Dot2+Dot3+Dot4",),
			"kb:q": ("br(braillesense):Dot1+Dot2+Dot3+Dot4+Dot5",),
			"kb:r": ("br(braillesense):Dot1+Dot2+Dot3+Dot5",),
			"kb:s": ("br(braillesense):Dot2+Dot3+Dot4",),
			"kb:t": ("br(braillesense):Dot2+Dot3+Dot4+Dot5",),
			"kb:u": ("br(braillesense):Dot1+Dot3+Dot6",),
			"kb:v": ("br(braillesense):Dot1+Dot2+Dot3+Dot6",),
			"kb:w": ("br(braillesense):Dot2+Dot4+Dot5+Dot6",),
			"kb:x": ("br(braillesense):Dot1+Dot3+Dot4+Dot6",),
			"kb:y": ("br(braillesense):Dot1+Dot3+Dot4+Dot5+Dot6",),
			"kb:z": ("br(braillesense):Dot1+Dot3+Dot5+Dot6",),
			"kb:,": ("br(braillesense):Dot2",),
			"kb:.": ("br(braillesense):Dot3",),
			"kb:-": ("br(braillesense):Dot3+Dot6",),
			"kb:;": ("br(braillesense):Dot2+Dot3",),
			"kb::": ("br(braillesense):Dot2+Dot5",),
			"kb:_": ("br(braillesense):Dot4+Dot5+Dot6",),
			"kb:!": ("br(braillesense):Dot5",),
			"kb:+": ("br(braillesense):Dot2+Dot3+Dot5",),
			"kb:*": ("br(braillesense):Dot3+Dot5",),
			"kb:'": ("br(braillesense):Dot6",),
			"kb:/": ("br(braillesense):Dot2+Dot5+Dot6",),
			"kb:(": ("br(braillesense):Dot2+Dot3+Dot6",),
			"kb:)": ("br(braillesense):Dot3+Dot5+Dot6",),
			'kb:"': ("br(braillesense):Dot4",),
			"kb:1": ("br(braillesense):Dot1+Dot6",),
			"kb:2": ("br(braillesense):Dot1+Dot2+Dot6",),
			"kb:3": ("br(braillesense):Dot1+Dot4+Dot6",),
			"kb:4": ("br(braillesense):Dot1+Dot4+Dot5+Dot6",),
			"kb:5": ("br(braillesense):Dot1+Dot5+Dot6",),
			"kb:6": ("br(braillesense):Dot1+Dot2+Dot4+Dot6",),
			"kb:7": ("br(braillesense):Dot1+Dot2+Dot4+Dot5+Dot6",),
			"kb:8": ("br(braillesense):Dot1+Dot2+Dot5+Dot6",),
			"kb:9": ("br(braillesense):Dot2+Dot4+Dot6",),
			"kb:0": ("br(braillesense):Dot3+Dot4+Dot6",),
			"kb:windows+b": ("br(braillesense):Dot1+Dot2+advance1",),
			"kb:windows+d": ("br(braillesense):Dot1+Dot4+Dot5+advance1",),
			"braille_routeTo": ("br(braillesense):routing",),
			"braille_scrollBack": ("br(braillesense):scroll_left",),
			"braille_scrollForward": ("br(braillesense):scroll_right",),

		}

	})


class InputGesture(braille.BrailleDisplayGesture):
	source = BrailleDisplayDriver.name
	def __init__(self, keys):
		global BS_KEYS
		super(InputGesture, self).__init__()
		if isinstance(keys,int):  
			self.routingIndex = keys
			self.id = "routing"
			return
		self.keyCodes = set(keys)
		names = set()
		for name in self.keyCodes: 
			try:
				names.add(BS_KEYS[name])
			except:
				pass
		self.id = "+".join(names)



